"""
**Extraction API Examples**

**Extract All Data for a Ticker**

.. code-block:: python

    import analysis_engine.extract as ae_extract
    print(ae_extract.extract('SPY'))

**Extract Latest Minute Pricing for Stocks and Options**

.. code-block:: python

    import analysis_engine.extract as ae_extract
    print(ae_extract.extract(
        'SPY',
        datasets=['minute', 'tdcalls', 'tdputs']))

**Extract Historical Data**

Extract historical data with the ``date`` argument formatted ``YYYY-MM-DD``:

.. code-block:: python

    import analysis_engine.extract as ae_extract
    print(ae_extract.extract(
        'AAPL',
        datasets=['minute', 'daily', 'financials', 'earnings', 'dividends'],
        date='2019-02-15'))

**Additional Extraction APIs**

`IEX Cloud Extraction API Reference <https://stock-analysis-engine.
readthedocs.io/en/latest/iex_api.html#iex-extraction-api-reference>`__

`Tradier Extraction API Reference <https://stock-analysis-engine.
readthedocs.io/en/latest/tradier.html#tradier-extraction-api-reference>`__

"""

import os
import analysis_engine.consts as ae_consts
import analysis_engine.utils as ae_utils
import analysis_engine.build_dataset_node as build_dataset_node
import analysis_engine.api_requests as api_requests
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def extract(
        ticker=None,
        datasets=None,
        tickers=None,
        use_key=None,
        extract_mode='all',
        iex_datasets=None,
        date=None,
        redis_enabled=True,
        redis_address=None,
        redis_db=None,
        redis_password=None,
        redis_expire=None,
        s3_enabled=True,
        s3_address=None,
        s3_bucket=None,
        s3_access_key=None,
        s3_secret_key=None,
        s3_region_name=None,
        s3_secure=False,
        celery_disabled=True,
        broker_url=None,
        result_backend=None,
        label=None,
        verbose=False):
    """extract

    Extract all cached datasets for a stock ``ticker`` or
    a list of ``tickers`` and returns a dictionary. Please
    make sure the datasets are already cached in Redis
    before running this method. If not please refer to
    the ``analysis_engine.fetch.fetch`` function
    to prepare the datasets on your environment.

    Python example:

    .. code-block:: python

        from analysis_engine.extract import extract
        d = extract(ticker='NFLX')
        print(d)
        for k in d['NFLX']:
            print(f'dataset key: {k}')

    **Extract Intraday Stock and Options Minute Pricing Data**

    This works by using the ``date`` and ``datasets`` arguments
    as filters:

    .. code-block:: python

        import analysis_engine.extract as ae_extract
        print(ae_extract.extract(
            ticker='SPY',
            datasets=['minute', 'tdcalls', 'tdputs'])

    This was created for reducing the amount of typying in
    Jupyter notebooks. It can be set up for use with a
    distributed engine as well with the optional arguments
    depending on your connectitivty requirements.

    .. note:: Please ensure Redis and Minio are running
              before trying to extract tickers

    **Stock tickers to extract**

    :param ticker: single stock ticker/symbol/ETF to extract
    :param tickers: optional - list of tickers to extract
    :param use_key: optional - extract historical key from Redis
        usually formatted ``<TICKER>_<date formatted YYYY-MM-DD>``

    **(Optional) Data sources, datafeeds and datasets to gather**

    :param iex_datasets: list of strings for gathering specific `IEX
        datasets <https://iexcloud.io/>`__
        which are set as consts: ``analysis_engine.iex.consts.FETCH_*``.
    :param date: optional - string date formatted
        ``YYYY-MM-DD`` - if not set use last close date
    :param datasets: list of strings for indicator
        dataset extraction - preferred method
        (defaults to ``BACKUP_DATASETS``)

    **(Optional) Redis connectivity arguments**

    :param redis_enabled: bool - toggle for auto-caching all
        datasets in Redis
        (default is ``True``)
    :param redis_address: Redis connection string format: ``host:port``
        (default is ``localhost:6379``)
    :param redis_db: Redis db to use
        (default is ``0``)
    :param redis_password: optional - Redis password
        (default is ``None``)
    :param redis_expire: optional - Redis expire value
        (default is ``None``)

    **(Optional) Minio (S3) connectivity arguments**

    :param s3_enabled: bool - toggle for auto-archiving on Minio (S3)
        (default is ``True``)
    :param s3_address: Minio S3 connection string format: ``host:port``
        (default is ``localhost:9000``)
    :param s3_bucket: S3 Bucket for storing the artifacts
        (default is ``dev``) which should be viewable on a browser:
        http://localhost:9000/minio/dev/
    :param s3_access_key: S3 Access key
        (default is ``trexaccesskey``)
    :param s3_secret_key: S3 Secret key
        (default is ``trex123321``)
    :param s3_region_name: S3 region name
        (default is ``us-east-1``)
    :param s3_secure: Transmit using tls encryption
        (default is ``False``)

    **(Optional) Celery worker broker connectivity arguments**

    :param celery_disabled: bool - toggle synchronous mode or publish
        to an engine connected to the `Celery broker and backend
        <https://github.com/celery/celery#transports-and-backends>`__
        (default is ``True`` - synchronous mode without an engine
        or need for a broker or backend for Celery)
    :param broker_url: Celery broker url
        (default is ``redis://0.0.0.0:6379/13``)
    :param result_backend: Celery backend url
        (default is ``redis://0.0.0.0:6379/14``)
    :param label: tracking log label

    **(Optional) Debugging**

    :param verbose: bool - show extract warnings
        and other debug logging (default is False)

    **Supported environment variables**

    ::

        export REDIS_ADDRESS="localhost:6379"
        export REDIS_DB="0"
        export S3_ADDRESS="localhost:9000"
        export S3_BUCKET="dev"
        export AWS_ACCESS_KEY_ID="trexaccesskey"
        export AWS_SECRET_ACCESS_KEY="trex123321"
        export AWS_DEFAULT_REGION="us-east-1"
        export S3_SECURE="0"
        export WORKER_BROKER_URL="redis://0.0.0.0:6379/13"
        export WORKER_BACKEND_URL="redis://0.0.0.0:6379/14"
    """

    rec = {}
    extract_requests = []

    use_tickers = tickers
    if ticker:
        use_tickers = [ticker]
    else:
        if not use_tickers:
            use_tickers = []

    default_iex_datasets = [
        'daily',
        'minute',
        'quote',
        'stats',
        'peers',
        'news',
        'financials',
        'earnings',
        'dividends',
        'company'
    ]

    if not iex_datasets:
        iex_datasets = default_iex_datasets

    use_indicator_datasets = datasets
    if not use_indicator_datasets:
        use_indicator_datasets = ae_consts.BACKUP_DATASETS

    if redis_enabled:
        if not redis_address:
            redis_address = os.getenv(
                'REDIS_ADDRESS',
                'localhost:6379')
        if not redis_password:
            redis_password = os.getenv(
                'REDIS_PASSWORD',
                None)
        if not redis_db:
            redis_db = int(os.getenv(
                'REDIS_DB',
                '0'))
        if not redis_expire:
            redis_expire = os.getenv(
                'REDIS_EXPIRE',
                None)
    if s3_enabled:
        if not s3_address:
            s3_address = os.getenv(
                'S3_ADDRESS',
                'localhost:9000')
        if not s3_access_key:
            s3_access_key = os.getenv(
                'AWS_ACCESS_KEY_ID',
                'trexaccesskey')
        if not s3_secret_key:
            s3_secret_key = os.getenv(
                'AWS_SECRET_ACCESS_KEY',
                'trex123321')
        if not s3_region_name:
            s3_region_name = os.getenv(
                'AWS_DEFAULT_REGION',
                'us-east-1')
        if not s3_secure:
            s3_secure = os.getenv(
                'S3_SECURE',
                '0') == '1'
        if not s3_bucket:
            s3_bucket = os.getenv(
                'S3_BUCKET',
                'dev')
    if not broker_url:
        broker_url = os.getenv(
            'WORKER_BROKER_URL',
            'redis://0.0.0.0:6379/13')
    if not result_backend:
        result_backend = os.getenv(
            'WORKER_BACKEND_URL',
            'redis://0.0.0.0:6379/14')

    if not label:
        label = 'get-latest'

    last_close_str = ae_utils.get_last_close_str()
    use_date_str = last_close_str
    if date:
        use_date_str = date

    ticker_key = use_key
    if use_key:
        ticker_key = f'{ticker}_{last_close_str}'
    else:
        ticker_key = f'{ticker}_{use_date_str}'

    common_vals = {}
    common_vals['base_key'] = ticker_key
    common_vals['celery_disabled'] = celery_disabled
    common_vals['ticker'] = ticker
    common_vals['label'] = label
    common_vals['iex_datasets'] = iex_datasets
    common_vals['s3_enabled'] = s3_enabled
    common_vals['s3_bucket'] = s3_bucket
    common_vals['s3_address'] = s3_address
    common_vals['s3_secure'] = s3_secure
    common_vals['s3_region_name'] = s3_region_name
    common_vals['s3_access_key'] = s3_access_key
    common_vals['s3_secret_key'] = s3_secret_key
    common_vals['s3_key'] = ticker_key
    common_vals['redis_enabled'] = redis_enabled
    common_vals['redis_address'] = redis_address
    common_vals['redis_password'] = redis_password
    common_vals['redis_db'] = redis_db
    common_vals['redis_key'] = ticker_key
    common_vals['redis_expire'] = redis_expire

    common_vals['redis_address'] = redis_address
    common_vals['s3_address'] = s3_address

    if verbose:
        log.info(
            f'{label} - extract ticker={ticker} last_close={last_close_str} '
            f'base_key={common_vals["base_key"]} '
            f'redis_address={common_vals["redis_address"]} '
            f's3_address={common_vals["s3_address"]}')

    """
    Extract Supported Datasets
    """

    for ticker in use_tickers:
        req = api_requests.get_ds_dict(
            ticker=ticker,
            base_key=common_vals['base_key'],
            ds_id=label)
        extract_requests.append(req)
    # end of for all ticker in use_tickers

    for extract_req in extract_requests:
        ticker_data = build_dataset_node.build_dataset_node(
            ticker=ticker,
            date=use_date_str,
            datasets=use_indicator_datasets,
            verbose=verbose)

        rec[ticker] = ticker_data
    # end of for service_dict in extract_requests

    return rec
# end of extract



#!/usr/bin/env python

"""
Tool for inspecting cached pricing data to find common errors.
This tool uses the
`Extraction API <https://stock-analysis-engine.
readthedocs.io/en/latest/extract.html>`__ to look for dates
that are not in sync with the redis cached date.

.. note:: This tool requires redis to be running with
    fetched datasets already stored in supported
    keys

**Examples**

**Inspect Minute Datasets for a Ticker**

::

    inspect_datasets.py -t SPY

**Inspect Daily Datasets for a Ticker**

::

    inspect_datasets.py -t AAPL -g daily
    # or
    # inspect_datasets.py -t AAPL -g day

**Usage**

::

    inspect_datasets.py -h
    usage: inspect_datasets.py [-h] [-t TICKER] [-g DATASETS] [-s START_DATE]

    Inspect datasets looking for dates in redis that look incorrect

    optional arguments:
    -h, --help     show this help message and exit
    -t TICKER      ticker
    -g DATASETS    optional - datasets: minute or min = examine IEX Cloud
                    intraday minute data, daily or day = examine IEX Cloud
                    daily
                    data, quote = examine IEX Cloud quotes data, stats =
                    examine
                    IEX Cloud key stats data, peers = examine IEX Cloud
                    peers
                    data, news = examine IEX Cloud news data, fin = examine
                    IEX
                    Cloud financials data, earn = examine IEX Cloud earnings
                    data, div = examine IEX Cloud dividendsdata, comp =
                    examine
                    IEX Cloud company data, calls = examine Tradier calls
                    data,
                    puts = examine Tradier puts data, and comma delimited is
                    supported as well
    -s START_DATE  start date format YYYY-MM-DD (default is 2019-01-01)
"""

import datetime
import argparse
import analysis_engine.consts as ae_consts
import analysis_engine.utils as ae_utils
import analysis_engine.extract as ae_extract
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(
    name='inspect-redis-data',
    handler_name='no_date_colors')


def inspect_datasets(
        ticker=None,
        start_date=None,
        datasets=None):
    """inspect_datasets

    Loop over all cached data in redis by going sequentially per date
    and examine the latest ``date`` value in the cache to
    check if it matches the redis key's date.

    For IEX Cloud minute data errors, running this function will print out
    commands to fix any issues (if possible):

    ::

        fetch -t TICKER -g iex_min -F DATE_TO_FIX

    :param ticker: optional - string ticker
    :param start_date: optional - datetime
        start date for the loop
        (default is ``2019-01-01``)
    :param datasets: optional - list of strings
        to extract specific, supported datasets
        (default is ``['minute']``)
    """

    if not start_date:
        start_date = datetime.datetime(
            year=2019,
            month=1,
            day=1)
    if not datasets:
        datasets = [
            'minute'
        ]
    if not ticker:
        ticker = 'SPY'

    tickers = [
        ticker
    ]

    fix_suggestions = []
    last_close = ae_utils.last_close()
    for ticker in tickers:

        not_done = True
        cur_date = start_date
        while not_done:
            cur_date_str = cur_date.strftime(ae_consts.COMMON_DATE_FORMAT)

            log.info(
                f'extracting {ticker} date={cur_date_str}')

            res = None

            # get from a date or the latest if not set
            if cur_date_str:
                res = ae_extract.extract(
                    ticker=ticker,
                    date=cur_date_str,
                    datasets=datasets)
            else:
                res = ae_extract.extract(
                    ticker=ticker,
                    datasets=datasets)

            weekday_name = cur_date.strftime('%A')

            for ds_name in datasets:
                df = res[ticker][ds_name]

                if ae_consts.is_df(df=df):
                    if 'date' in df:
                        latest_date = df['date'].iloc[-1]
                        latest_date_str = latest_date.strftime(
                            ae_consts.COMMON_DATE_FORMAT)
                        if latest_date_str == cur_date_str:
                            log.info(
                                f'valid - {ds_name} latest dates match '
                                f'{weekday_name}: '
                                f'{latest_date_str} == {cur_date_str}')
                        else:
                            if ds_name != 'daily':
                                log.critical(
                                    f'{ds_name} latest dates does '
                                    f'NOT match on '
                                    f'{weekday_name} {cur_date_str} found: '
                                    f'{latest_date_str}')
                            else:
                                one_day_back = (
                                    latest_date + datetime.timedelta(days=1))
                                if weekday_name == 'Monday':
                                    one_day_back = (
                                        latest_date + datetime.timedelta(
                                            days=3))
                                latest_date_str = one_day_back.strftime(
                                    ae_consts.COMMON_DATE_FORMAT)
                                if latest_date_str == cur_date_str:
                                    log.info(
                                        f'valid - {ds_name} latest dates '
                                        f'match '
                                        f'{weekday_name}: '
                                        f'{latest_date_str} == '
                                        f'{cur_date_str}')
                                else:
                                    log.critical(
                                        f'{ds_name} latest dates does '
                                        f'NOT match on '
                                        f'{weekday_name} {cur_date_str} '
                                        f'found: '
                                        f'{latest_date_str}')

                            if ds_name == 'minute':
                                fix_suggestions.append(
                                    f'fetch -t {ticker} -g iex_min '
                                    f'-F {cur_date_str}')
                    else:
                        log.error(
                            f'{ds_name} df does not have a date column '
                            f'on {cur_date_str}')
                else:
                    log.error(
                        f'Missing {ds_name} df on {cur_date_str}')
            # end of inspecting datasets

            if cur_date > last_close:
                not_done = False
            else:
                cur_date += datetime.timedelta(days=1)
                not_a_weekday = True
                while not_a_weekday:
                    weekday = cur_date.date().weekday()
                    if weekday > 4:
                        log.debug(
                            'SKIP weekend day: '
                            f'{cur_date.strftime("%A on %Y-%m-%d")}')
                        cur_date += datetime.timedelta(days=1)
                    else:
                        not_a_weekday = False
        # end for all dates
    # end of for all tickers

    if len(fix_suggestions) > 0:
        print('-------------------------------')
        print(
            'Detected invalid dates - below are the suggested fixes '
            'to run using the fetch command.')
        print(
            ' - Please be aware fetching data may incur usages and '
            'costs on your account')
        for s in fix_suggestions:
            print(s)
    else:
        log.info(
            'done')
# end inspect_datasets


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description=(
            'Inspect datasets looking for dates in redis '
            'that look incorrect'))
    parser.add_argument(
        '-t',
        help=(
            'ticker'),
        required=False,
        dest='ticker')
    parser.add_argument(
        '-g',
        help=(
            'optional - datasets: '
            'minute or min = examine IEX Cloud intraday minute data, '
            'daily or day  = examine IEX Cloud daily data, '
            'quote = examine IEX Cloud quotes data, '
            'stats = examine IEX Cloud key stats data, '
            'peers = examine IEX Cloud peers data, '
            'news = examine IEX Cloud news data, '
            'fin = examine IEX Cloud financials data, '
            'earn = examine IEX Cloud earnings data, '
            'div = examine IEX Cloud dividendsdata, '
            'comp = examine IEX Cloud company data, '
            'calls = examine Tradier calls data, '
            'puts = examine Tradier puts data, '
            'and comma delimited is supported as well'),
        required=False,
        dest='datasets')
    parser.add_argument(
        '-s',
        help=(
            'start date format YYYY-MM-DD (default is 2019-01-01)'),
        required=False,
        dest='start_date')
    args = parser.parse_args()

    start_date = datetime.datetime(
        year=2019,
        month=1,
        day=1)
    datasets = [
        'minute'
    ]
    ticker = 'SPY'

    valid = True
    if args.ticker:
        ticker = args.ticker.upper()
    if args.datasets:
        datasets = []
        for key in args.datasets.lower().split(','):
            if key == 'news':
                datasets.append('news1')
            elif key == 'min':
                datasets.append('minute')
            elif key == 'day':
                datasets.append('daily')
            elif key == 'fin':
                datasets.append('financials')
            elif key == 'earn':
                datasets.append('earnings')
            elif key == 'div':
                datasets.append('dividends')
            elif key == 'comp':
                datasets.append('company')
            elif key == 'calls':
                datasets.append('tdcalls')
            elif key == 'puts':
                datasets.append('tdputs')
            else:
                if key not in ae_consts.BACKUP_DATASETS:
                    log.error(
                        f'unsupported dataset key: {key} '
                        'please use a supported key: '
                        f'{ae_consts.BACKUP_DATASETS}')
                    valid = False
                else:
                    datasets.append(key)
    if args.start_date:
        start_date = datetime.datetime.strptime(
            args.start_date,
            '%Y-%m-%d')

    if valid:
        inspect_datasets(
            ticker=ticker,
            start_date=start_date,
            datasets=datasets)



"""
Fetch API calls for pulling IEX Cloud Data from
a valid IEX account

.. warning:: Running these API calls will impact
    your account's monthly quota. Please be
    aware of your usage when calling these.

Please set the environment variable ``IEX_TOKEN`` to
your account token before running these calls.

More steps can be found on the docs in the
`IEX API <https://stock-analysis-engine.readth
edocs.io/en/latest/iex_api.html#iex-api>`__

**Command Line Tool Fetching Examples**

With the Analysis Engine stack running you can use
the pip's included ``fetch`` command line tool with the
following arguments to pull data (and automate it).

**Fetch Minute Data**

::

    fetch -t AAPL -g min

**Fetch Daily Data**

::

    fetch -t AAPL -g day

**Fetch Quote Data**

::

    fetch -t AAPL -g quote

**Fetch Stats Data**

::

    fetch -t AAPL -g stats

**Fetch Peers Data**

::

    fetch -t AAPL -g peers

**Fetch News Data**

::

    fetch -t AAPL -g news

**Fetch Financials Data**

::

    fetch -t AAPL -g fin

**Fetch Earnings Data**

::

    fetch -t AAPL -g earn

**Fetch Dividends Data**

::

    fetch -t AAPL -g div

**Fetch Company Data**

::

    fetch -t AAPL -g comp

**Command Line Fetch Debugging**

Add the ``-d`` flag to the ``fetch`` command to enable
verbose logging. Here is an example:

::

    fetch -t AAPL -g news -d

"""

import pandas as pd
import analysis_engine.consts as ae_consts
import analysis_engine.utils as ae_utils
import analysis_engine.dataset_scrub_utils as dataset_utils
import analysis_engine.iex.consts as iex_consts
import analysis_engine.iex.helpers_for_iex_api as iex_helpers
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def fetch_daily(
        ticker=None,
        work_dict=None,
        scrub_mode='sort-by-date',
        verbose=False):
    """fetch_daily

    Fetch the IEX daily data for a ticker and
    return it as a ``pandas.DataFrame``.

    https://iexcloud.io/docs/api/#historical-prices

    .. code-block:: python

        import analysis_engine.iex.fetch_api as iex_fetch

        daily_df = iex_fetch.fetch_daily(ticker='SPY')
        print(daily_df)

    :param ticker: string ticker to fetch
    :param work_dict: dictionary of args
        used by the automation
    :param scrub_mode: optional - string
        type of scrubbing handler to run
    :param verbose: optional - bool to log for debugging
    """
    label = None
    if work_dict:
        if not ticker:
            ticker = work_dict.get('ticker', None)
        label = work_dict.get('label', None)

    use_url = (
        f'/stock/{ticker}/chart/1m')

    if verbose:
        log.info(
            f'{label} - daily - url={use_url} '
            f'req={work_dict} '
            f'ticker={ticker}')

    resp_json = iex_helpers.get_from_iex(
        url=use_url,
        token=iex_consts.IEX_TOKEN,
        verbose=verbose)

    df = pd.DataFrame(resp_json)

    if verbose:
        log.info(
            f'{label} - daily - url={use_url} '
            f'ticker={ticker} response '
            f'df={df.tail(5)}')

    if len(df.index) == 0:
        return df

    iex_helpers.convert_datetime_columns(
        df=df)

    cols_to_drop = [
    ]

    remove_these = None
    for c in df:
        if c in cols_to_drop:
            if not remove_these:
                remove_these = []
            remove_these.append(c)

    if remove_these:
        df = df.drop(columns=remove_these)

    df.set_index(
        [
            'date'
        ]).sort_values(by=['date'], ascending=True)

    return df
# end of fetch_daily


def fetch_minute(
        ticker=None,
        backfill_date=None,
        work_dict=None,
        scrub_mode='sort-by-date',
        verbose=False):
    """fetch_minute

    Fetch the IEX minute intraday data for a ticker and
    return it as a ``pandas.DataFrame``.

    https://iexcloud.io/docs/api/#historical-prices

    .. code-block:: python

        import analysis_engine.iex.fetch_api as iex_fetch

        minute_df = iex_fetch.fetch_minute(ticker='SPY')
        print(minute_df)

    :param ticker: string ticker to fetch
    :param backfill_date: optional - date string formatted
        ``YYYY-MM-DD`` for filling in missing minute data
    :param work_dict: dictionary of args
        used by the automation
    :param scrub_mode: optional - string
        type of scrubbing handler to run
    :param verbose: optional - bool to log for debugging
    """
    label = None
    use_date = backfill_date
    from_historical_date = None
    last_close_to_use = None
    dates = []

    if work_dict:
        label = work_dict.get('label', None)
        use_date = work_dict.get('use_date', None)
        if not ticker:
            ticker = work_dict.get('ticker', None)
        if not backfill_date:
            use_date = work_dict.get('backfill_date', None)
        if 'from_historical_date' in work_dict:
            from_historical_date = work_dict['from_historical_date']
        if 'last_close_to_use' in work_dict:
            last_close_to_use = work_dict['last_close_to_use']
        if from_historical_date and last_close_to_use:
            dates = ae_utils.get_days_between_dates(
                from_historical_date=work_dict['from_historical_date'],
                last_close_to_use=last_close_to_use)

    use_url = (
        f'/stock/{ticker}/chart/1d')

    if use_date:
        # no - chars in the date
        use_url = (
            f'/stock/{ticker}/chart/date/{use_date.replace("-", "")}')

    if verbose:
        log.info(
            f'{label} - minute - url={use_url} '
            f'req={work_dict} ticker={ticker} '
            f'fhdate={from_historical_date} '
            f'last_close={last_close_to_use} '
            f'dates={dates}')

    resp_json = iex_helpers.get_from_iex(
        url=use_url,
        token=iex_consts.IEX_TOKEN,
        verbose=verbose)

    df = pd.DataFrame(resp_json)

    if verbose:
        log.info(
            f'{label} - minute - url={use_url} '
            f'ticker={ticker} response '
            f'df={df.tail(5)}')

    if 'date' not in df:
        log.error(
            f'unable to download IEX Cloud minute '
            f'data for {ticker} on backfill_date={use_date} '
            f'df: {df} from url: {use_url} with response: {resp_json}')
        return df

    if len(df.index) == 0:
        return df

    iex_helpers.convert_datetime_columns(
        df=df)

    if not use_date:
        use_date = df['date'].iloc[-1].strftime('%Y-%m-%d')

    new_minutes = dataset_utils.build_dates_from_df_col(
        src_col='minute',
        use_date_str=use_date,
        df=df)
    df['date'] = pd.to_datetime(
        new_minutes,
        format=ae_consts.COMMON_TICK_DATE_FORMAT)
    # make sure dates are set as strings in the cache
    df['date'] = df['date'].dt.strftime(
        ae_consts.COMMON_TICK_DATE_FORMAT)
    df.set_index(
        [
            'date'
        ])
    return df
# end of fetch_minute


def fetch_quote(
        ticker=None,
        work_dict=None,
        scrub_mode='sort-by-date',
        verbose=False):
    """fetch_quote

    Fetch the IEX quote data for a ticker and
    return as a ``pandas.DataFrame``.

    https://iexcloud.io/docs/api/#quote

    .. code-block:: python

        import analysis_engine.iex.fetch_api as iex_fetch

        quote_df = iex_fetch.fetch_quote(ticker='SPY')
        print(quote_df)

    :param ticker: string ticker to fetch
    :param work_dict: dictionary of args
        used by the automation
    :param scrub_mode: optional - string
        type of scrubbing handler to run
    :param verbose: optional - bool to log for debugging
    """
    label = None
    if work_dict:
        if not ticker:
            ticker = work_dict.get('ticker', None)
        label = work_dict.get('label', None)

    use_url = (
        f'/stock/{ticker}/quote')

    if verbose:
        log.info(
            f'{label} - quote - url={use_url} '
            f'req={work_dict} ticker={ticker}')

    resp_json = iex_helpers.get_from_iex(
        url=use_url,
        token=iex_consts.IEX_TOKEN,
        verbose=verbose)

    df = pd.DataFrame([resp_json])

    if verbose:
        log.info(
            f'{label} - quote - url={use_url} '
            f'ticker={ticker} response '
            f'df={df.tail(5)}')

    if len(df.index) == 0:
        return df

    iex_helpers.convert_datetime_columns(
        df=df)

    cols_to_drop = []
    remove_these = None
    if len(cols_to_drop) > 0:
        for c in df:
            if c in cols_to_drop:
                if not remove_these:
                    remove_these = []
                remove_these.append(c)

    if remove_these:
        df = df.drop(columns=remove_these)

    return df
# end of fetch_quote


def fetch_stats(
        ticker=None,
        work_dict=None,
        scrub_mode='sort-by-date',
        verbose=False):
    """fetch_stats

    Fetch the IEX statistics data for a ticker and
    return it as a ``pandas.DataFrame``.

    https://iexcloud.io/docs/api/#key-stats

    .. code-block:: python

        import analysis_engine.iex.fetch_api as iex_fetch

        stats_df = iex_fetch.fetch_stats(ticker='SPY')
        print(stats_df)

    :param ticker: string ticker to fetch
    :param work_dict: dictionary of args
        used by the automation
    :param scrub_mode: optional - string
        type of scrubbing handler to run
    :param verbose: optional - bool to log for debugging
    """
    label = None
    if work_dict:
        if not ticker:
            ticker = work_dict.get('ticker', None)
        label = work_dict.get('label', None)

    use_url = (
        f'/stock/{ticker}/stats')

    if verbose:
        log.info(
            f'{label} - stats - url={use_url} '
            f'req={work_dict} ticker={ticker}')

    resp_json = iex_helpers.get_from_iex(
        url=use_url,
        token=iex_consts.IEX_TOKEN,
        verbose=verbose)

    df = pd.DataFrame([resp_json])

    if verbose:
        log.info(
            f'{label} - stats - url={use_url} '
            f'ticker={ticker} response '
            f'df={df.tail(5)}')

    if len(df.index) == 0:
        return df

    iex_helpers.convert_datetime_columns(
        df=df)

    cols_to_drop = []
    remove_these = None
    if len(cols_to_drop) > 0:
        for c in df:
            if c in cols_to_drop:
                if not remove_these:
                    remove_these = []
                remove_these.append(c)

    if remove_these:
        df = df.drop(columns=remove_these)

    return df
# end of fetch_stats


def fetch_peers(
        ticker=None,
        work_dict=None,
        scrub_mode='sort-by-date',
        verbose=False):
    """fetch_peers

    Fetch the IEX peers data for a ticker and
    return it as a ``pandas.DataFrame``.

    https://iexcloud.io/docs/api/#peers

    .. code-block:: python

        import analysis_engine.iex.fetch_api as iex_fetch

        peers_df = iex_fetch.fetch_peers(ticker='SPY')
        print(peers_df)

    :param ticker: string ticker to fetch
    :param work_dict: dictionary of args
        used by the automation
    :param scrub_mode: optional - string
        type of scrubbing handler to run
    :param verbose: optional - bool to log for debugging
    """
    label = None
    if work_dict:
        if not ticker:
            ticker = work_dict.get('ticker', None)
        label = work_dict.get('label', None)

    use_url = (
        f'/stock/{ticker}/relevant')

    if verbose:
        log.info(
            f'{label} - peers - url={use_url} '
            f'req={work_dict} ticker={ticker}')

    resp_json = iex_helpers.get_from_iex(
        url=use_url,
        token=iex_consts.IEX_TOKEN,
        verbose=verbose)

    df = pd.DataFrame(resp_json)

    if verbose:
        log.info(
            f'{label} - peers - url={use_url} '
            f'ticker={ticker} response '
            f'df={df.tail(5)}')

    if len(df.index) == 0:
        return df

    iex_helpers.convert_datetime_columns(
        df=df)

    cols_to_drop = []
    remove_these = None
    if len(cols_to_drop) > 0:
        for c in df:
            if c in cols_to_drop:
                if not remove_these:
                    remove_these = []
                remove_these.append(c)

    if remove_these:
        df = df.drop(columns=remove_these)

    return df
# end of fetch_peers


def fetch_news(
        ticker=None,
        num_news=5,
        work_dict=None,
        scrub_mode='sort-by-date',
        verbose=False):
    """fetch_news

    Fetch the IEX news data for a ticker and
    return it as a ``pandas.DataFrame``.

    https://iexcloud.io/docs/api/#news

    .. code-block:: python

        import analysis_engine.iex.fetch_api as iex_fetch

        news_df = iex_fetch.fetch_news(ticker='SPY')
        print(news_df)

    :param ticker: string ticker to fetch
    :param num_news: optional - int number of news
        articles to fetch
        (default is ``5`` articles)
    :param work_dict: dictionary of args
        used by the automation
    :param scrub_mode: optional - string
        type of scrubbing handler to run
    :param verbose: optional - bool to log for debugging
    """
    label = None
    if work_dict:
        if not ticker:
            ticker = work_dict.get('ticker', None)
        label = work_dict.get('label', None)
        if not num_news:
            num_news = int(work_dict.get('num_news', 5))

    use_url = (
        f'/stock/{ticker}/news/last/{num_news}')

    if verbose:
        log.info(
            f'{label} - news - url={use_url} '
            f'req={work_dict} ticker={ticker}')

    resp_json = iex_helpers.get_from_iex(
        url=use_url,
        token=iex_consts.IEX_TOKEN,
        verbose=verbose)

    df = pd.DataFrame(resp_json)

    if verbose:
        log.info(
            f'{label} - news - url={use_url} '
            f'ticker={ticker} response '
            f'df={df.tail(5)}')

    if len(df.index) == 0:
        return df

    if 'datetime' not in df:
        log.error(
            f'unable to download IEX Cloud news '
            f'data for {ticker} '
            f'df: {df} from url: {use_url} with response: {resp_json}')
        return df

    df['datetime'] = pd.to_datetime(
        df['datetime'],
        unit='ms',
        errors='coerce')

    cols_to_drop = []
    remove_these = None
    if len(cols_to_drop) > 0:
        for c in df:
            if c in cols_to_drop:
                if not remove_these:
                    remove_these = []
                remove_these.append(c)

    if remove_these:
        df = df.drop(columns=remove_these)

    return df
# end of fetch_news


def fetch_financials(
        ticker=None,
        work_dict=None,
        scrub_mode='sort-by-date',
        verbose=False):
    """fetch_financials

    Fetch the IEX financial data for a ticker and
    return it as a ``pandas.DataFrame``.

    https://iexcloud.io/docs/api/#financials

    .. code-block:: python

        import analysis_engine.iex.fetch_api as iex_fetch

        fin_df = iex_fetch.fetch_financials(ticker='SPY')
        print(fin_df)

    :param ticker: string ticker to fetch
    :param work_dict: dictionary of args
        used by the automation
    :param scrub_mode: optional - string
        type of scrubbing handler to run
    :param verbose: optional - bool to log for debugging
    """
    label = None
    if work_dict:
        if not ticker:
            ticker = work_dict.get('ticker', None)
        label = work_dict.get('label', None)

    use_url = (
        f'/stock/{ticker}/financials')

    if verbose:
        log.info(
            f'{label} - fins - url={use_url} '
            f'req={work_dict} ticker={ticker}')

    resp_json = iex_helpers.get_from_iex(
        url=use_url,
        token=iex_consts.IEX_TOKEN,
        verbose=verbose)

    df = pd.DataFrame(resp_json.get('financials', []))

    if verbose:
        log.info(
            f'{label} - fins - url={use_url} '
            f'ticker={ticker} response '
            f'df={df.tail(5)}')

    if len(df.index) == 0:
        return df

    iex_helpers.convert_datetime_columns(
        df=df)

    cols_to_drop = []
    remove_these = None
    if len(cols_to_drop) > 0:
        for c in df:
            if c in cols_to_drop:
                if not remove_these:
                    remove_these = []
                remove_these.append(c)

    if remove_these:
        df = df.drop(columns=remove_these)

    return df
# end of fetch_financials


def fetch_earnings(
        ticker=None,
        work_dict=None,
        scrub_mode='sort-by-date',
        verbose=False):
    """fetch_earnings

    Fetch the IEX earnings data for a ticker and
    return it as a ``pandas.DataFrame``.

    https://iexcloud.io/docs/api/#earnings

    .. code-block:: python

        import analysis_engine.iex.fetch_api as iex_fetch

        earn_df = iex_fetch.fetch_earnings(ticker='SPY')
        print(earn_df)

    :param ticker: string ticker to fetch
    :param work_dict: dictionary of args
        used by the automation
    :param scrub_mode: optional - string
        type of scrubbing handler to run
    :param verbose: optional - bool to log for debugging
    """
    label = None
    if work_dict:
        if not ticker:
            ticker = work_dict.get('ticker', None)
        label = work_dict.get('label', None)

    use_url = (
        f'/stock/{ticker}/earnings')

    if verbose:
        log.info(
            f'{label} - earns - url={use_url} '
            f'req={work_dict} ticker={ticker}')

    resp_json = iex_helpers.get_from_iex(
        url=use_url,
        token=iex_consts.IEX_TOKEN,
        verbose=verbose)

    df = pd.DataFrame(resp_json.get('earnings', []))

    if verbose:
        log.info(
            f'{label} - earns - url={use_url} '
            f'ticker={ticker} response '
            f'df={df.tail(5)}')

    if len(df.index) == 0:
        return df

    iex_helpers.convert_datetime_columns(
        df=df)

    cols_to_drop = []
    remove_these = None
    if len(cols_to_drop) > 0:
        for c in df:
            if c in cols_to_drop:
                if not remove_these:
                    remove_these = []
                remove_these.append(c)

    if remove_these:
        df = df.drop(columns=remove_these)

    return df
# end of fetch_earnings


def fetch_dividends(
        ticker=None,
        timeframe='3m',
        work_dict=None,
        scrub_mode='sort-by-date',
        verbose=False):
    """fetch_dividends

    Fetch the IEX dividends data for a ticker and
    return it as a ``pandas.DataFrame``.

    https://iexcloud.io/docs/api/#dividends

    .. code-block:: python

        import analysis_engine.iex.fetch_api as iex_fetch

        div_df = iex_fetch.fetch_dividends(ticker='SPY')
        print(div_df)

    :param ticker: string ticker to fetch
    :param timeframe: optional - string for setting
        dividend lookback period used for
        (default is ``3m`` for three months)
    :param work_dict: dictionary of args
        used by the automation
    :param scrub_mode: optional - string
        type of scrubbing handler to run
    :param verbose: optional - bool to log for debugging
    """
    label = None
    if work_dict:
        if not ticker:
            ticker = work_dict.get('ticker', None)
        label = work_dict.get('label', None)
        if not timeframe:
            timeframe = work_dict.get('timeframe', '3m')

    use_url = (
        f'/stock/{ticker}/dividends/{timeframe}')

    if verbose:
        log.info(
            f'{label} - divs - url={use_url} '
            f'req={work_dict} ticker={ticker}')

    resp_json = iex_helpers.get_from_iex(
        url=use_url,
        token=iex_consts.IEX_TOKEN,
        verbose=verbose)

    df = pd.DataFrame(resp_json)

    if verbose:
        log.info(
            f'{label} - divs - url={use_url} '
            f'ticker={ticker} response '
            f'df={df.tail(5)}')

    if len(df.index) == 0:
        return df

    iex_helpers.convert_datetime_columns(
        df=df)

    cols_to_drop = []
    remove_these = None
    if len(cols_to_drop) > 0:
        for c in df:
            if c in cols_to_drop:
                if not remove_these:
                    remove_these = []
                remove_these.append(c)

    if remove_these:
        df = df.drop(columns=remove_these)

    return df
# end of fetch_dividends


def fetch_company(
        ticker=None,
        work_dict=None,
        scrub_mode='NO_SORT',
        verbose=False):
    """fetch_company

    Fetch the IEX company data for a ticker and
    return it as a ``pandas.DataFrame``.

    https://iexcloud.io/docs/api/#company

    .. code-block:: python

        import analysis_engine.iex.fetch_api as iex_fetch

        comp_df = iex_fetch.fetch_company(ticker='SPY')
        print(comp_df)

    :param ticker: string ticker to fetch
    :param work_dict: dictionary of args
        used by the automation
    :param scrub_mode: optional - string
        type of scrubbing handler to run
    :param verbose: optional - bool to log for debugging
    """
    label = None
    if work_dict:
        if not ticker:
            ticker = work_dict.get('ticker', None)
        label = work_dict.get('label', None)

    use_url = (
        f'/stock/{ticker}/company')

    if verbose:
        log.info(
            f'{label} - comp - url={use_url} '
            f'req={work_dict} ticker={ticker}')

    resp_json = iex_helpers.get_from_iex(
        url=use_url,
        token=iex_consts.IEX_TOKEN,
        verbose=verbose)

    df = pd.DataFrame([resp_json])

    if verbose:
        log.info(
            f'{label} - comp - url={use_url} '
            f'ticker={ticker} response '
            f'df={df}')

    if len(df.index) == 0:
        return df

    iex_helpers.convert_datetime_columns(
        df=df)

    cols_to_drop = []
    remove_these = None
    if len(cols_to_drop) > 0:
        for c in df:
            if c in cols_to_drop:
                if not remove_these:
                    remove_these = []
                remove_these.append(c)

    if remove_these:
        df = df.drop(columns=remove_these)

    return df
# end of fetch_company
